@@ -30,12 +30,13 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <QMutexLocker>
-#include <QThreadStorage>
 #include <QTimer>
+#include <QBitmap>
+#include <QPixmap>
+#include <QCursor>
 
 //for detecting intel AMT KVM vnc server
-static const QString INTEL_AMT_KVM_STRING= QLatin1String("Intel(r) AMT KVM");
-static QThreadStorage<VncClientThread **> instances;
+static const QString INTEL_AMT_KVM_STRING = QLatin1String("Intel(r) AMT KVM");
 
 // Dispatch from this static callback context to the member context.
 rfbBool VncClientThread::newclientStatic(rfbClient *cl)
@@ -67,7 +68,7 @@ void VncClientThread::cuttextStatic(rfbClient *cl, const char *text, int textlen
 // Dispatch from this static callback context to the member context.
 char *VncClientThread::passwdHandlerStatic(rfbClient *cl)
 {
-    VncClientThread *t = (VncClientThread *)rfbClientGetClientData(cl, nullptr);
+    VncClientThread *t = (VncClientThread *) rfbClientGetClientData(cl, nullptr);
     Q_ASSERT(t);
 
     return t->passwdHandler();
@@ -76,7 +77,7 @@ char *VncClientThread::passwdHandlerStatic(rfbClient *cl)
 // Dispatch from this static callback context to the member context.
 rfbCredential *VncClientThread::credentialHandlerStatic(rfbClient *cl, int credentialType)
 {
-    VncClientThread *t = (VncClientThread *)rfbClientGetClientData(cl, nullptr);
+    VncClientThread *t = (VncClientThread *) rfbClientGetClientData(cl, nullptr);
     Q_ASSERT(t);
 
     return t->credentialHandler(credentialType);
@@ -85,21 +86,56 @@ rfbCredential *VncClientThread::credentialHandlerStatic(rfbClient *cl, int crede
 // Dispatch from this static callback context to the member context.
 void VncClientThread::outputHandlerStatic(const char *format, ...)
 {
-    VncClientThread **t = instances.localData();
+    VncClientThread *t = qobject_cast<VncClientThread *>(QThread::currentThread());
+    Q_ASSERT(t);
 
     va_list args;
     va_start(args, format);
-    (*t)->outputHandler(format, args);
+    t->outputHandler(format, args);
     va_end(args);
 }
 
+void VncClientThread::cursorShapeHandlerStatic(rfbClient *cl, int xhot, int yhot, int width, int height, int bpp) {
+    VncClientThread *t = (VncClientThread *) rfbClientGetClientData(cl, nullptr);
+    Q_ASSERT(t);
+
+    // get cursor shape from remote cursor field
+    // it's important to set stride for images, pictures in VNC are not always 32-bit aligned
+    QImage cursorImg;
+    switch (bpp) {
+    case 4:
+        cursorImg = QImage(cl->rcSource, width, height, bpp * width, QImage::Format_RGB32);
+        break;
+    case 2:
+        cursorImg = QImage(cl->rcSource, width, height, bpp * width, QImage::Format_RGB16);
+        break;
+    case 1:
+        cursorImg = QImage(cl->rcSource, width, height, bpp * width, QImage::Format_Indexed8);
+        cursorImg.setColorTable(t->m_colorTable);
+        break;
+    default:
+        qCWarning(KRDC) << "Unsupported bpp value for cursor shape:" << bpp;
+        return;
+    }
+
+    // get alpha channel
+    QImage alpha(cl->rcMask, width, height, 1 * width, QImage::Format_Indexed8);
+    alpha.setColorTable({qRgb(255, 255, 255), qRgb(0, 0, 0)});
+
+    // apply transparency mask
+    QPixmap cursorPixmap(QPixmap::fromImage(cursorImg));
+    cursorPixmap.setMask(QBitmap::fromImage(alpha));
+
+    emit t->gotCursor({cursorPixmap, xhot, yhot});
+}
+
 void VncClientThread::setClientColorDepth(rfbClient* cl, VncClientThread::ColorDepth cd)
 {
     switch(cd) {
     case bpp8:
         if (m_colorTable.isEmpty()) {
             m_colorTable.resize(256);
-            int r,g,b;
+            int r, g, b;
             for (int i = 0; i < 256; ++i) {
                 //pick out the red (3 bits), green (3 bits) and blue (2 bits) bits and make them maximum significant in 8bits
                 //this gives a colortable for 8bit true colors
@@ -194,10 +230,10 @@ void VncClientThread::updatefb(int x, int y, int w, int h)
         img.setColorTable(m_colorTable);
         break;
     case bpp16:
-        img = QImage(cl->frameBuffer, width, height, 2*width, QImage::Format_RGB16);
+        img = QImage(cl->frameBuffer, width, height, 2 * width, QImage::Format_RGB16);
         break;
     case bpp32:
-        img = QImage(cl->frameBuffer, width, height, 4*width, QImage::Format_RGB32);
+        img = QImage(cl->frameBuffer, width, height, 4 * width, QImage::Format_RGB32);
         break;
     }
 
@@ -209,6 +245,7 @@ void VncClientThread::updatefb(int x, int y, int w, int h)
         return; // sending data to a stopped thread is not a good idea
     }
 
+    img.setDevicePixelRatio(m_devicePixelRatio);
     setImage(img);
 
     emitUpdated(x, y, w, h);
@@ -323,6 +360,7 @@ VncClientThread::VncClientThread(QObject *parent)
         : QThread(parent)
         , frameBuffer(nullptr)
         , cl(nullptr)
+        , m_devicePixelRatio(1.0)
         , m_stopped(false)
 {
     // We choose a small value for interval...after all if the connection is
@@ -380,6 +418,23 @@ void VncClientThread::setPort(int port)
     m_port = port;
 }
 
+void VncClientThread::setShowLocalCursor(bool show) {
+    QMutexLocker locker(&mutex);
+    m_showLocalCursor = show;
+
+    if (!cl) {
+        // no client yet, only store local value
+        return;
+    }
+
+    // from server point of view, "remote" cursor is the one local to the client
+    // so the meaning in AppData struct is inverted
+    cl->appData.useRemoteCursor = show;
+
+    // need to communicate this change to server or it won't stop painting cursor
+    m_eventQueue.enqueue(new ReconfigureEvent);
+}
+
 void VncClientThread::setQuality(RemoteView::Quality quality)
 {
     m_quality = quality;
@@ -397,9 +452,14 @@ void VncClientThread::setQuality(RemoteView::Quality quality)
     }
 }
 
+void VncClientThread::setDevicePixelRatio(qreal dpr)
+{
+    m_devicePixelRatio = dpr;
+}
+
 void VncClientThread::setColorDepth(ColorDepth colorDepth)
 {
-    m_colorDepth= colorDepth;
+    m_colorDepth = colorDepth;
 }
 
 RemoteView::Quality VncClientThread::quality() const
@@ -448,9 +508,6 @@ void VncClientThread::run()
 {
     QMutexLocker locker(&mutex);
 
-    VncClientThread **threadTls = new VncClientThread *();
-    *threadTls = this;
-    instances.setLocalData(threadTls);
     while (!m_stopped) { // try to connect as long as the server allows
         locker.relock();
         m_passwordError = false;
@@ -535,8 +592,11 @@ bool VncClientThread::clientCreate(bool reinitialising)
     cl->GetCredential = credentialHandlerStatic;
     cl->GotFrameBufferUpdate = updatefbStatic;
     cl->GotXCutText = cuttextStatic;
+    cl->GotCursorShape = cursorShapeHandlerStatic;
     rfbClientSetClientData(cl, nullptr, this);
 
+    cl->appData.useRemoteCursor = m_showLocalCursor;
+
     cl->serverHost = strdup(m_host.toUtf8().constData());
 
     cl->serverPort = m_port;
@@ -631,6 +691,11 @@ ClientEvent::~ClientEvent()
 {
 }
 
+void ReconfigureEvent::fire(rfbClient* cl)
+{
+    SetFormatAndEncodings(cl);
+}
+
 void PointerClientEvent::fire(rfbClient* cl)
 {
     SendPointerEvent(cl, m_x, m_y, m_buttonMask);
